import math
import matplotlib.pyplot as plt
import random
import scipy.stats as calculator


def plot_my_dict(my_dict, y_axis):
    colors = dict[int:str]()
    for h in my_dict:
        colors.update({h: "#" + ''.join([random.choice('0123456789ABCDEF') for j in range(6)])})

    plt.style.use('_mpl-gallery')

    # make data
    for h in my_dict:
        y = my_dict[h]
        x = range(1, 30)

        plt.plot(x, y, color=colors[h], label=h)

    plt.xlabel("c")
    plt.ylabel(y_axis)
    plt.legend(loc="upper right")
    plt.show()

def visualising_1(length_h):
    '''
    Exercise 1: visualises the false positive rate of bloom filters with c of 1 to 30
    :param length_h: nuber of different hash functions
    :return: the FPR standard
    '''
    values = dict[int:list]()
    for h in range(length_h):
        values.update({h: []})
        for c in range(1, 30):
            values[h].append(pow(base=(1-math.exp(-h/c)), exp=h))

    plot_my_dict(values, "FPR")
    return values


def calc_overall_FPR_512(h, c):
    '''
    THe calculation of all the different values for the Blocked FPR
    :param h:  the h for which to compute
    :param c:  the c for which to compute
    :return: returns the overall FPR for block Bloom of size 512
    '''
    k = 1
    fpr = 0
    poiss= 1
    while poiss > 0.001:
        poiss = calculator.poisson.pmf(k, 512 / c)

        binom = calculator.binom.pmf(k, 512, pow(base=(511/512), exp=h*k))
        print(binom)
        fpr_z = 0
        for z in range(513):
            fpr_z += (1-z/512)**h

        fpr += fpr_z*binom*poiss
        k += 1
        poiss = calculator.poisson.pmf(k, 512 / c)

    return fpr


def visualising_2(length_h):
    '''
    Exercise 2: Calculating the FPR for Blocked Bloom filter of size 512
    :param length_h: the max value of h for which to calculate
    :return: returns all block FPR values
    '''
    values = dict[int:list]()
    for h in range(length_h):
        values.update({h: []})
        for c in range(1, 30):
            values[h].append(calc_overall_FPR_512(h, c))

    plot_my_dict(values, "FPR")
    return values


def comparison_of_FPR_values(standard, blocked):
    '''
    Exercise 3: comparison of the standard and the block bloom filter FPR ratio
    :param standard: dictionary of the standard FPR
    :param blocked: dictionary of the block FPR
    :return: nothing but happiness
    '''
    R = dict[int:list]()
    for h in standard:
        R.update({h: []})
        for c in range(30):
            R[h].append(blocked[h][c]/standard[h][c])

        # make data
    plot_my_dict(R, "Ratio of FPR of Block/standard")


def main():
    '''
    The function to run if you want everything to happen
    :return: nothing
    '''
    #standard = visualising_1(10)
    blocked = visualising_2(10)
    #comparison_of_FPR_values(standard, blocked)


if __name__ == "__main__":
    main()