import this
from math import ceil
from collections import namedtuple

from math import prod

import numpy as np
from numba import njit, int64, uint64



bitarray = namedtuple("BitArray", [
    "size",
    "array",       # !
    "get",         # !
    "set",         # !
    ])

def BitArray(size):

    def aligned_zeros(shape, *, dtype=np.uint64, byte_alignment=64, autogrow=False):
        """
        Allocate and return a buffer of zeros of the given `shape` (int or tuple) and `dtype`.
        Ensure that the byte-address of the buffer is divisible by `byte_alignment` bytes.
        A cache line typically has 512 bits = 64 bytes,
        so byte_alignment=64 ensures that the buffer starts at a chache line boundary.
        This may waste a few bytes
        (with dtype=np.uint64 and byte_alignment=64 up to 56 = 64 - 8).
        """

        if byte_alignment % 8 != 0:
            raise ValueError("aligned_zeros: byte_alignment must be a multiple of 8 bytes")
        one_dimensional = isinstance(shape, int) or len(shape) == 1
        items = shape if isinstance(shape, int) else prod(shape)
        itemsize = np.dtype(dtype).itemsize
        items_per_cacheline = 64 // itemsize  # 64 bytes per cache line
        if not one_dimensional:
            # check that buffer uses a multiple of 64 bytes in lowest dimension
            if (hanging := (shape[-1] % items_per_cacheline)) != 0:
                if autogrow:
                    shape = tuple(shape[:-1]) + (shape[-1] + items_per_cacheline - hanging, )
                    items = prod(shape)
                else:
                    raise ValueError(f"aligned_zeros: last dimension has {hanging} overhanging items (not aligned)")
        slack = 64 - itemsize
        buf = np.zeros(items + slack, dtype=dtype)
        address = buf.ctypes.data  # buf.__array_interface__['data'][0]
        al = address % byte_alignment
        shift = ((byte_alignment - al) // itemsize) if al != 0 else 0
        # shift start of buffer by 'shift' items to achieve the requested alignment
        assert 0 <= shift < slack
        b = buf[shift:(shift + items)]
        baddress = b.ctypes.data  # b.__array_interface__['data'][0]
        if items > 0:
            assert baddress % byte_alignment == 0, f"{byte_alignment=}, {items=}, {itemsize=}, {slack=}, {al=}, {shift=}, {address=}, {baddress=}"
        return b if one_dimensional else b.reshape(*shape)

    def intbitarray(n, w):
        """
        Initialize and return a bitarray of 'size' bits.
        Ensure that the first element is aligned to an 'alignment'-byte address,
        e.g. use alignment=64 (bytes) for 512-bit alignment (cache line size).

        Alternatively, pass an existing uint64[:] arary as the 'size' parameter
        in order to interpret it as bit array.

        """
        w = max(1, w)
        w = min(64, w)
        n = max(1, n)

        size = ceil(n*w/64)

        if isinstance(size, np.ndarray):
            array = size
            if array.dtype != np.uint64:
                raise TypeError(f"bitarray: dtype of given array is {array.dtype}, but must be uint64")
            size = int(array.size * 64)  # size in bits of existing array
        else:
            array = None
            size = int(size)

        array = aligned_zeros(size)


    def get(a, i):
        return get_old(a, i*w, w)

    @njit(nogil=True, locals=dict(
          start=int64, x=uint64, mask=uint64, mask1=uint64, bits=int64))
    def get_old(a, start, bits=1):
        """return 'bits' bits from a[start:start+bits], where bits <= 64"""
        if bits <= 0:
            return uint64(0)
        startint = start // 64  # item starts in a[startint]
        startbit = start & 63   # at bit number startbit
        if startbit + bits <= 64:
            # bits are contained in a single uint64
            x = a[startint]
            if startbit > 0:
                x >>= startbit
        else:
            # bits are distributed over two uint64s,
            # less significant bits are the leftmost b1=(64-startbit) bits in a[startint]
            # more significant bits are the rightmost (bits-64+startbit) bits in a[startint+1]
            b1 = 64 - startbit
            mask1 = uint64(uint64(1 << b1) - uint64(1))
            x = uint64(a[startint] >> startbit) & mask1 
            x |= uint64(a[startint + 1] << b1)
        # due to a bug in numba, do not use x = y if cond else z !!
        if bits >= 64:
            return x
        mask = uint64(2**bits - 1)
        x &= mask
        return x


    def set(a, i, value):
        set_old(a, i*w, value, w)
    @njit(nogil=True, locals=dict(
          start=int64, value=uint64, v1=uint64,
          mask=uint64, mask1=uint64, mask2=uint64))
    def set_old(a, start, value, bits=1):
        """set a[start:start+bits] to value, where bits <= 64"""
        if bits <= 0:
            return
        startint = start // 64  # item starts in a[startint]
        startbit = start & 63   # at bit number startbit
        if bits >= 64:
            mask = uint64(-1)
        else:
            mask = uint64(2**bits - 1)
        if startbit + bits <= 64:
            # bits are contained in a single uint64
            mask1 = ~(mask << startbit)
            a[startint] = (a[startint] & mask1) | (value << startbit)
        else:
            # b1 leftmost bits in a[startint] == b1 rightmost bits of v, 
            b1 = 64 - startbit
            v1 = (value & uint64(2**b1 - 1))  # v1 = b1 rightmost bits of v
            mask1 = uint64(2**startbit - 1)  # only keep startbit rightmost bits
            a[startint] = (a[startint] & mask1) | (v1 << startbit)
            # b2 rightmost bits in a[startint+1] = b2 leftmost bits of v
            b2 = bits - b1
            mask2 = uint64(~(2**b2 - 1))
            a[startint + 1] = (a[startint + 1] & mask2) | (value >> b1)


